/***********************************************************************
* E.S.O. - VLT project
*
* who       when      what
* --------  --------  ----------------------------------------------
* gzins     24/04/03  Created
* gzins     28/04/03  Documented
* gzins     24/05/06  added a print of amdlib library version.
*/

/**
 * @file
 * Program to convert AMBER binary table data files to FITS files.
 *
 * @synopsis
 * amdlibBtbl2Fits <btblFile1> [<btblFile2> ...]
 *
 * @param btblFileX name of files containing binary tables
 *
 * @details
 * This program creates FITS files from data files produced by the AMBER
 * instrument which are stored as binary tables. This consists to load the
 * original data file, to create glued images from data of detector regions and
 * save them into FITS file. The basename of the FITS files is the basename of
 * the binary files with _IMG. The file extension is still '.fits'. For
 * example, the binary file AMBER_CAL_0001.fits will be converted into a FITS
 * file named AMBER_CAL_0001_IMG.fits.
 */

static char *rcsId="@(#) $Id: amdlibBtbl2Fits.c,v 1.5 2011/01/24 12:43:31 agabasch Exp $"; 
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

#define _POSIX_SOURCE 1


/*
 * System Headers 
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <strings.h>

/*
 * Local Headers 
 */
#include "amdlib.h"
#include "amdlibProtected.h"

/*
 * Local Functions
 */
static void amdlibPrintUsage(const char *programName)
{
    printf(" Usage: %s [-h] [-v level] <btblFile1> [<btblFile2> ...]\n", 
           programName);
    printf("  Options:\n");
    printf("    -h help\n");
    printf("    -v verbosity level [0..4]\n");
}


/**
 * @internal
 * Main program
 */
int main (int argc, char *argv[])
{
    extern int optind;
    extern char *optarg;
    int c;
    int nbOfParams;
    const char *programName;
    int level;
    int i;
    char fitsFileName[256];
    char *chrPtr;

    
    /* Print library version */
    amdlibPrintVersion();

    /* Get program name */
    programName = amdlibGetProgramName(argv[0]);

    /* Start program */
    optind = 1;
    while (1)
    {
        c = getopt(argc, argv, "hv:");
        if (c == -1)
        {
            break;
        }
        switch (c)
        {
            case 'v':
                if (sscanf (optarg, "%i", &level) == 0)
                {
                    amdlibLogError("Invalid numerical value for \'-v\' "
                                   "option");
                    return EXIT_FAILURE;
                }
                if ((level < amdlibLOG_QUIET) || (level > amdlibLOG_TRACE))
                {
                    amdlibLogError("Invalid  value for \'-v\' option");
                    amdlibLogErrorDetail("Valid range is [%d..%d]", 
                                         amdlibLOG_QUIET, amdlibLOG_TRACE);
                    return EXIT_FAILURE;
                }
                amdlibLogSet(amdlibLogIsEnabled(), level, amdlibFALSE);
                break;
            case 'h':
                amdlibPrintUsage(programName);
                return EXIT_SUCCESS;
            default:
                amdlibPrintUsage(programName);
                return EXIT_FAILURE;
        }
    }
    
    /* Check number of parameters */
    /* number of parameters = argc (all arguments) - optind (contains program
     * name + options) */
    nbOfParams = argc - optind;
    if (nbOfParams == 0)
    {
        amdlibLogError("%s %s", amdlibERR_NB_PARAM, programName);
        amdlibPrintUsage(programName);
        return EXIT_FAILURE;
    }
   
    /* For all input binary table file */
    for (i = optind; i < argc; i++)
    {
        memset(fitsFileName, '\0', sizeof(fitsFileName));
        strncpy(fitsFileName, argv[i], sizeof(fitsFileName) - 9);
        /* Cats down the '.fits' extention from the filename. */
        if ((chrPtr = strstr(fitsFileName, ".fits")) != NULL)
        {
            *chrPtr = '\0';
        }
        /* Adds '_IMG.fits' */
        strcat(fitsFileName, "_IMG.fits");

        /* And then, convert to FITS file */
        amdlibLogInfo("Converting %s to %s...", argv[i], fitsFileName);
        if (amdlibBtbl2Fits(argv[i], fitsFileName) != amdlibSUCCESS)
        {
            amdlibLogError("Could not convert file '%s' to FITS file", 
                           argv[i]);
            return EXIT_FAILURE;
        }
    }

    return EXIT_SUCCESS;
}

/*___oOo___*/
