/************************************************************************
 * E.S.O. - VLT project
 *
 * who       when      what
 * --------  --------  ----------------------------------------------
 * altariba  12/21/05  Created (extracted from amdlibVisibilities.c)
 * ealtariba 05/24/06  Added treatement about frame selection.
 * gduvert   07/06/06  modified error computations, simplified writing
 *                     and set phases in radians.
 */

/**
 * @file
 * Useful functions related to closure phases.
 */

#define _POSIX_SOURCE 1

static char *rcsId="@(#) $Id: amdlibClosurePhases.c,v 1.5 2011/01/24 12:43:33 agabasch Exp $";
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

/*
 * System Headers
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "fitsio.h"

/*
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"


/** Useful macro to free all dynamically allocated structures */ 
#define amdlibComputeClosurePhases_FREEALL()       \
    amdlibFree2DArrayWrapping((void **)tablePtr);  \
    amdlibFree2DArrayWrapping((void **)table3Ptr); 
/**
 * Compute closure phases. 
 *
 * This function computes the closure phases, averaged over a bin,
 * from a bin of instantaneous correlated fluxes. 
 * 
 * @param instantCorrFlux instantaneous bin of correlated flux.
 * @param selectedFrames structure containing information relative to frame
 * selection. By default, all frames are selected.
 * @param vis3 data structure (binned) where results are stored.
 * @param iBin index of the result in the output vis3 structure.
 * @param errMsg error description message returned if function fails.
 *
 * @return 
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */

amdlibCOMPL_STAT amdlibComputeClosurePhases(/* Input */
                                            amdlibVIS       *instantCorrFlux,
                                            int             iBin,
                                            amdlibBAND      band,
                                            amdlibSELECTION *selectedFrames,
                                            amdlibERROR_TYPE   errorType,
                                            /* Output */
                                            amdlibVIS3      *vis3,
                                            amdlibERROR_MSG errMsg)
{
    double Re_bispectrum = 0.0;
    double Im_bispectrum = 0.0;
    double Sigma2_bispectrum_re = 0.0;
    double Sigma2_bispectrum_im = 0.0;
    double Re2_bispectrum = 0.0;
    double Im2_bispectrum = 0.0;
    double Re4Im4_bispectrum = 0.0;
    double x, y, s2x, s2y, x1, x2, x3, y1, y2, y3;
    double s2x1, s2x2, s2x3, s2y1, s2y2, s2y3;
    double val;

    int iClos;
    int lVis;
    int nbClos   = vis3->nbClosures;
    int nbWlen   = vis3->nbWlen;
    int nbBases  = instantCorrFlux->nbBases;
    int nbOkFrames = selectedFrames->band[band].nbFramesOkForClosure;

    int iFrame, i;
    amdlibVIS_TABLE_ENTRY **tablePtr = NULL;
    amdlibVIS3_TABLE_ENTRY **table3Ptr = NULL;

    amdlibLogTrace("amdlibComputeClosurePhases()");

    /* Wrap instantCorrFlux->table structure to easy the reading of the code*/
    tablePtr = 
        (amdlibVIS_TABLE_ENTRY **)amdlibWrap2DArray(instantCorrFlux->table,
                                                nbBases, 
                                                instantCorrFlux->nbFrames,
                                                sizeof(amdlibVIS_TABLE_ENTRY),
                                                errMsg);
    if (tablePtr == NULL)
    {
        amdlibComputeClosurePhases_FREEALL();
        return amdlibFAILURE;
    }
    
    table3Ptr = 
        (amdlibVIS3_TABLE_ENTRY **)amdlibWrap2DArray(vis3->table,
                                                nbClos, vis3->nbFrames,
                                                sizeof(amdlibVIS3_TABLE_ENTRY),
                                                errMsg);
    if (table3Ptr == NULL)
    {
        amdlibComputeClosurePhases_FREEALL();
        return amdlibFAILURE;
    }

    for (iClos = 0; iClos < nbClos; iClos++) 
    {
        for (lVis = 0; lVis < nbWlen; lVis++)
        { /* begin loop for lVis */

            Re_bispectrum = 0.0;
            Im_bispectrum = 0.0;
            Re2_bispectrum = 0.0;
            Im2_bispectrum = 0.0;
            Re4Im4_bispectrum = 0.0;
            Sigma2_bispectrum_re = 0.0;
            Sigma2_bispectrum_im = 0.0;
            for (i = 0; i < nbOkFrames; i++)
            {
                iFrame = selectedFrames->band[band].frameOkForClosure[i];
                x1 = tablePtr[iFrame][0].vis[lVis].re;
                x2 = tablePtr[iFrame][1].vis[lVis].re;
                x3 = tablePtr[iFrame][2].vis[lVis].re;
                y1 = tablePtr[iFrame][0].vis[lVis].im;
                y2 = tablePtr[iFrame][1].vis[lVis].im;
                y3 = tablePtr[iFrame][2].vis[lVis].im;
                s2x1 = tablePtr[iFrame][0].sigma2Vis[lVis].re;
                s2x2 = tablePtr[iFrame][1].sigma2Vis[lVis].re;
                s2x3 = tablePtr[iFrame][2].sigma2Vis[lVis].re;
                s2y1 = tablePtr[iFrame][0].sigma2Vis[lVis].im;
                s2y2 = tablePtr[iFrame][1].sigma2Vis[lVis].im;
                s2y3 = tablePtr[iFrame][2].sigma2Vis[lVis].im;
		val = x1*x2*x3 - y1*y2*x3 + y1*x2*y3 + x1*y2*y3;
                Re_bispectrum += val;
                Re2_bispectrum += amdlibPow2(val);
                Re4Im4_bispectrum += amdlibPow2(val*val);
                val = y1*y2*y3 - x1*x2*y3 + x1*y2*x3 + y1*x2*x3;
                Im_bispectrum += val;
                Im2_bispectrum += amdlibPow2(val);
                Re4Im4_bispectrum += amdlibPow2(val*val);

                if ((errorType == amdlibTHEORETICAL_ERROR)||
                    (errorType == amdlibSTATISTICAL_ERROR))
                    /* PLEASE MODIFY BELOW TO ADD PROPER 
                     * STATISTICAL ERROR TRETAMENT (action FM) 
                     */
                {
                    /* Please have a look to the F. Millour thesis
                       (http://tel.archives-ouvertes.fr/tel-00134268),
                       pp.84-85 (eq. 4.34 and 4.35) */
                    Sigma2_bispectrum_re +=
                        s2x1 * (amdlibPow2(x2*x3) + amdlibPow2(y2*y3)) +
                        s2x2 * (amdlibPow2(x1*x3) + amdlibPow2(y1*y3)) + 
                        s2x3 * (amdlibPow2(x1*x2) + amdlibPow2(y1*y2)) +
                        s2y1 * (amdlibPow2(y2*x3) + amdlibPow2(x2*y3)) + 
                        s2y2 * (amdlibPow2(y1*x3) + amdlibPow2(x1*y3)) +
                        s2y3 * (amdlibPow2(y2*x1) + amdlibPow2(x2*y1));

                    Sigma2_bispectrum_im +=
                        s2y1 * (amdlibPow2(x2*x3) + amdlibPow2(y2*y3)) + 
                        s2y2 * (amdlibPow2(x1*x3) + amdlibPow2(y1*y3)) +
                        s2y3 * (amdlibPow2(y1*y2) + amdlibPow2(x1*x2)) +
                        s2x1 * (amdlibPow2(x2*y3) + amdlibPow2(y2*x3)) +
                        s2x2 * (amdlibPow2(x1*y3) + amdlibPow2(y1*x3)) + 
                        s2x3 * (amdlibPow2(y1*x2) + amdlibPow2(x1*y2));
                }
                else if(errorType == amdlibSTATISTICAL_ERROR)
                {
                    Sigma2_bispectrum_re += 0;
                    Sigma2_bispectrum_im += 0;
                }
            }

            if (nbOkFrames != 0)
            {
                Re_bispectrum /= nbOkFrames;
                Im_bispectrum /= nbOkFrames;
                Re2_bispectrum /= nbOkFrames;
                Im2_bispectrum /= nbOkFrames;
                Re4Im4_bispectrum /= nbOkFrames;
                Sigma2_bispectrum_re /= nbOkFrames;
                Sigma2_bispectrum_im /= nbOkFrames;
            }

            x = Re_bispectrum;
            y = Im_bispectrum;
            s2x = Sigma2_bispectrum_re;
            s2y = Sigma2_bispectrum_im;

            table3Ptr[iBin][iClos].vis3Amplitude[lVis] =
                sqrt(x*x + y*y);
            table3Ptr[iBin][iClos].vis3Phi[lVis] =
                atan2(y,x);

            table3Ptr[iBin][iClos].vis3AmplitudeError[lVis] = 
                1 / (x*x + y*y) *x*x*s2x + 
                1 / (x*x+y*y) *y*y*s2y;

            /* Please have a look to the F. Millour thesis
               (http://tel.archives-ouvertes.fr/tel-00134268),
               pp.55-57 (eq 3.28) and pp.84-85 (eq 4.36 where a typo error
               removed the M factor) */
            table3Ptr[iBin][iClos].vis3PhiError[lVis] =
            sqrt( (s2x * Im2_bispectrum + s2y * Re2_bispectrum) /
                  Re4Im4_bispectrum );
        }
    }

    amdlibComputeClosurePhases_FREEALL();
    return amdlibSUCCESS;
}
#undef amdlibComputeClosurePhases_FREEALL


/**
 * Computes a crude weighted average of all closures phase in a closure
 * phase structure. 
 *
 * @param vis3 input/output: closure structure, where averaged values are
 * writen in member "averageClosure".
 */
void amdlibAverageClosurePhases(amdlibVIS3 *vis3)
{
    double spectralClosurePhase[amdlibNB_SPECTRAL_CHANNELS];
    double spectralClosurePhaseErr[amdlibNB_SPECTRAL_CHANNELS];
    double averageClosurePhase;
    double averageClosurePhaseErr;
    double c,e;

    int iClos;
    int lVis;
    int iFrame;
    const int nbClos = 1; /* Number of Closures FIXED */
    int nbWlen =  vis3->nbWlen;
    int nbFrames = vis3->nbFrames;
    int nbGoodClos = 0;
    int nbGoodClosErr = 0;
    int nc = 0;

    amdlibLogTrace("amdlibAverageClosurePhases()");
    
    averageClosurePhase = 0;
    averageClosurePhaseErr = 0;
    amdlibVIS3_TABLE_ENTRY **table3Ptr = NULL;
    static amdlibERROR_MSG errMsg;

    table3Ptr = 
        (amdlibVIS3_TABLE_ENTRY **)amdlibWrap2DArray(vis3->table,
                                                nbClos, nbFrames,
                                                sizeof(amdlibVIS3_TABLE_ENTRY),
                                                errMsg);
    if (table3Ptr == NULL)
    {
        amdlibFree2DArrayWrapping((void **)table3Ptr);
    }
    
    for (lVis = 0; lVis < nbWlen; lVis++)
    {
        nc = 0;
        spectralClosurePhase[lVis] = 0;
        spectralClosurePhaseErr[lVis] = 0;
        for (iClos = 0; iClos < nbClos; iClos++)
        { 
            for (iFrame = 0; iFrame < nbFrames; iFrame++)
            { 
                c = table3Ptr[iFrame][iClos].vis3Phi[lVis];
                e = table3Ptr[iFrame][iClos].vis3PhiError[lVis];
                if (c == c && e == e) /* Trick to avoid NaN values */
                {
                    spectralClosurePhase[lVis] +=
                        table3Ptr[iFrame][iClos].vis3Phi[lVis];
                    spectralClosurePhaseErr[lVis] +=
                        table3Ptr[iFrame][iClos].vis3PhiError[lVis];
                    nc++;
                }
            }
        }
        spectralClosurePhase[lVis] /= nc;
        spectralClosurePhaseErr[lVis] /= nc;
        
        if(spectralClosurePhase[lVis] == spectralClosurePhase[lVis]) 
        {
            averageClosurePhase += spectralClosurePhase[lVis];
            nbGoodClos++;
        }
        if(spectralClosurePhaseErr[lVis] == spectralClosurePhaseErr[lVis]) 
        {
            averageClosurePhaseErr += spectralClosurePhaseErr[lVis];
            nbGoodClosErr++;
        }
    }
    if (nbGoodClos) 
    {
        vis3->averageClosure = averageClosurePhase / nbGoodClos;
    }
    if (nbGoodClosErr)
    {
        vis3->averageClosureError = averageClosurePhaseErr / 
                                    nbGoodClosErr;
    }
    amdlibFree2DArrayWrapping((void **)table3Ptr);
}


/*___oOo___*/
