/************************************************************************
 * E.S.O. - VLT project 
 *
 * who       when      what
 * --------  --------  ----------------------------------------------
 * gduvert   19/06/07  corrected mod 2*PI bug in amdlibComputeBaselines
 * gduvert   30/05/07  modified since sign of Piston is now reversed
 *                     to be consistent with the _wrong_ sign of the
 *                     PHASE method...
 * gduvert   28/05/07  protected amdlibRmsTable from nans and negative
 *                     sigma2 as well.
 * gduvert   25/05/07  Added Abacus for Phase error correction.  
 * gduvert   27/02/07  created amdlibComputeBaselines
 * gduvert   06/07/06  created and completed 
 *                     amdlibCorrect3DVisTableFromAchromaticPiston,
 *                     amdlibFindIndexOfMinimum, amdlibComputeMatrixCov and
 *                     amdlibComputeCov.
 * ealtariba 04/07/06  changed amdlibComputeCov function, added 
 *                     amdlibComputeMatrixCov function.
 *                     corrected bugs in amdlibAvgTable and 
 *                     amdlibRmsTable functions.
 *                     added amdlibCorrect3DVisTableFromAchromaticPiston
 *                     function.
 * gduvert   26/10/04  amdlibFindIndexOfMaximum: change *data from double to
 *                     float
 * gduvert   12/02/03  created
 */

/**
 * @file
 * Various (usefull) utilities for matrix manipulations
 */

#define _POSIX_SOURCE 1

static char *rcsId="@(#) $Id: amdlibMatrix.c,v 1.5 2011/01/24 12:43:44 agabasch Exp $"; 
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

/* 
 * System Headers
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "fitsio.h"
/* 
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"

/*
 * Protected functions
 */
/**
 * Invert squared matrix. 
 *
 * This function inverts a squared (dim,dim) matrix and writes the result in
 * place.
 *
 * @param matrix matrix to invert
 * @param dim dimension of the matrix
 *
 * @return
 * Always return amdlibSUCCESS
 */
amdlibCOMPL_STAT amdlibInvertMatrix(double *matrix, int dim)
{
    int    i;
    int    j;
    int    k;
    double sum = 0.0;
    double x = 1.0;

    amdlibLogTrace("amdlibInvertMatrix()");
    
    if (dim <= 0)
    {
        return amdlibFAILURE;  /* sanity check */
    }
    if (dim == 1) 
    {
        return amdlibFAILURE;  /* must be of dimension >= 2 */
    }

    for (i = 1; i < dim; i++)
    {
        matrix[i] /= matrix[0]; /* normalize row 0 */
    }

    for (i = 1; i < dim; i++)
    {
        for (j = i; j < dim; j++)  /* do a column of L */
        {
            sum = 0.0;
            for (k = 0; k < i; k++)
            {
                sum += matrix[j*dim + k] * matrix[k*dim + i];
            }
            matrix[j*dim + i] -= sum;
        }
        if (i == dim - 1)
        {
            continue;
        }
        for (j = i + 1; j < dim; j++)  /* do a row of U */
        {
            sum = 0.0;
            for (k = 0; k < i; k++)
            {
                sum += matrix[i*dim +k ] * matrix[k*dim + j];
            }
            matrix[i*dim + j] = (matrix[i*dim + j] - sum) / matrix[i*dim + i];
        }
    }
    for (i = 0; i < dim; i++)  /* invert L */
    {
        for (j = i; j < dim; j++)
        {
            x = 1.0;
            if (i != j)
            {
                x = 0.0;
                for (k = i; k < j; k++)
                    x -= matrix[j*dim + k] * matrix[k*dim + i];
            }
            matrix[j*dim + i] = x / matrix[j*dim + j];
        }
    }
    for (i = 0; i < dim; i++)   /* invert U */
    {
        for (j = i; j < dim; j++)
        {
            if (i == j)
            {
                continue;
            }
            sum = 0.0;
            for (k = i; k < j; k++)
            {
                sum += matrix[k*dim + j] * ((i == k) ? 1.0 : matrix[i*dim +k]);
            }
            matrix[i*dim + j] = -sum;
        }
    }

    for (i = 0; i < dim; i++)   /* final inversion */
    {
        for (j = 0; j < dim; j++)
        {
            sum = 0.0;
            for (k = ((i > j) ? i : j); k < dim; k++)
            {
                sum += ((j == k) ? 1.0 : matrix[j*dim + k]) * matrix[k*dim +i];
            }
            matrix[j*dim + i] = sum;
        }
    }
    return amdlibSUCCESS;
}

/**
 * Tranpose matrix. 
 *
 * This function transposes (dim1,dim2) matrix into (dim1,dim2) matrix 
 *
 * @param matrix matrix to transpose
 * @param tmatrix transposed  matrix
 * @param dim1 dimension along X-axis of the matrix
 * @param dim2 dimension along Y-axis of the matrix
 *
 * @return
 * Always return amdlibSUCCESS
 */
void amdlibTransposeMatrix(double *matrix, double *tmatrix, int dim1, int dim2)
{
    int i;
    int j;

    amdlibLogTrace("amdlibTransposeMatrix()");
    
    for (i = 0; i < dim2; i++)
    {
        for (j = 0; j < dim1; j++)
        {
            tmatrix[i*dim1 + j] = matrix[j*dim2 + i]; 
        }
    }
}

/**
 * Computes the product matrix
 *
 * This function computes the product of a (dim1,dim2) matrix with a (dim2,dim3)
 * matrix.
 *
 * @param matrix1 matrix 1 used in product 
 * @param matrix2 matrix 2 used in product  
 * @param matprod matrix resulting of matrix product
 * @param dim1 dimension along X-axis of the matrix 1
 * @param dim2 dimension along Y-axis of the matrix 1 and along X-axis of the
 * matrix 2
 * @param dim3 dimension along Y-axis of the matrix 2
 *
 * @return
 * Always return amdlibSUCCESS
 */
amdlibCOMPL_STAT amdlibProductMatrix(double *matrix1, double *matrix2,
                                     double *matprod, int dim1, int dim2,
                                     int dim3)
{
    int i,j,k;

    amdlibLogTrace("amdlibProductMatrix()");
    
    for (i = 0; i < dim1; i++)
    {
        for (j = 0; j < dim3; j++)
        {
            matprod[i*dim3 + j] = 0.;
            for (k = 0; k < dim2; k++)
            {
                matprod[i*dim3 + j] += matrix1[i*dim2 + k] * 
                    matrix2[k*dim3 + j];
            }
        }
    }
    return amdlibSUCCESS;
}

/**
 * Compute the covariance of 2 observables x and y
 *
 * This function computes the covariance of 2 observables x and y
 * matrixes. These observables are should be stored in multi-dimensional array
 * having the following dimensions [nbFrames][nbWlen][nbBases].
 * The resulting covariance is stored in a vector [nbWlen][nbBases].
 *
 * @param x first observable  
 * @param y second observable
 * @param lambda wavelength index at which covariance is computed
 * @param nbWlen number of wavelengths
 * @param nbFrames number of frames
 * @param nbBases number of bases
 * @param cov_xy vector where result is stored 
 */
void amdlibComputeMatrixCov(double *x, double *y, int lambda, int nbWlen, 
                      int nbFrames, int nbBases, double *cov_xy)
{
    int    kbase;
    int    ifram;
    double prod = 0.;
    double av_x = 0.;
    double av_y = 0.;

    amdlibLogTrace("amdlibComputeMatrixCov()");
    
    for (kbase = 0; kbase < nbBases; kbase++)
    {
        prod = 0.;
        av_x = 0.;
        av_y = 0.;
        for (ifram = 0; ifram < nbFrames; ifram++)
        {
            /* sum(xy) */
            prod += x[nbBases*nbWlen*ifram + nbBases*lambda + kbase] *
                y[nbBases*nbWlen*ifram + nbBases*lambda + kbase];
            /* sum(x) */
            av_x += x[nbBases*nbWlen*ifram + nbBases*lambda + kbase];
            /* sum(y) */
            av_y += y[nbBases*nbWlen*ifram + nbBases*lambda + kbase];
        }
        /* averages */
        prod /= nbFrames;
        av_x /= nbFrames;
        av_y /= nbFrames;
        /* cov(x,y) = average(xy)-average(x)*average(y) */
        cov_xy[nbBases*lambda + kbase] = prod - (av_x * av_y);
    }
}

/**
 * Compute the covariance of 2 vectors x and y
 *
 * This function computes the covariance of 2 vectors x and y.
 * having the dimension [nbFrames].
 * The resulting covariance is stored in a double.
 *
 * @param x first observable  
 * @param y second observable
 * @param nbFrames number of frames
 */
double amdlibComputeCov(double *x, double *y, int nbFrames)
{
    int    ifram;
    double prod = 0.;
    double av_x = 0.;
    double av_y = 0.;
    double cov_xy = 0.;

    amdlibLogTrace("amdlibComputeCov()");

    prod = 0.;
    av_x = 0.;
    av_y = 0.;
    for (ifram = 0; ifram < nbFrames; ifram++)
    {
        /* sum(xy) */
        prod += x[ifram] * y[ifram];
        /* sum(x) */
        av_x += x[ifram];
        /* sum(y) */
        av_y += y[ifram];
    }
    /* averages */
    prod /= nbFrames;
    av_x /= nbFrames;
    av_y /= nbFrames;
    /* cov(x,y) = average(xy)-average(x)*average(y) */
    cov_xy  = prod - (av_x * av_y);
    return (cov_xy);
}

/**
 * Calculate ponderated mean.
 *
 * @param nbPix size of table.
 * @param table input data.
 * @param sigma2 ponderation values.
 *
 * @return the average value.
 */
double amdlibAvgTable(int    nbPix,
                      double *table,
                      double *sigma2)
{
    double ponderation = 0;
    double avg = 0;
    int    iPix;

    for (iPix = 0; iPix < nbPix; iPix++)
    {
        if ((table[iPix] == table[iPix]) && (sigma2[iPix] > 0))
        {
            avg += table[iPix] / sigma2[iPix];
            ponderation += 1.0 / sigma2[iPix];
        }
    }
    if (ponderation > 0)
    {
        avg /= ponderation;
    }

    return avg;
}

/**
 * Calculate the residual rms.
 *
 * @param nbPix size of table.
 * @param table input data.
 * @param sigma2 ponderation values.
 *
 * @return the value of rms.
 */
double amdlibRmsTable(int    nbPix,
                      double *table,
                      double *sigma2)
{
    double ponderation = 0;
    double avg = 0;
    double rms = 0;
    int    iPix;

    avg = amdlibAvgTable(nbPix, table, sigma2);

    for (iPix = 0; iPix < nbPix; iPix++)
    {
        if ((table[iPix] == table[iPix]) && (sigma2[iPix] > 0))
        {
            rms += amdlibPow2(table[iPix] - avg) / sigma2[iPix];
            ponderation += 1.0 / sigma2[iPix];
        }
    }
    
    if (ponderation > 0)
    {
        rms /= ponderation;
        rms = sqrt(rms);
    }
    return rms;
}

/**
 * Compute signed square root
 *
 * @param a input value.
 *
 * @return the value computed.
 */
double amdlibSignedSqrt(double a)
{
    double val = a;
    if (a != 0.0)
    {
        val = (a / fabs(a)) * sqrt(fabs(a));
    }
    return val;
}

#define amdlibCorrect3DVisTableFromAchromaticPiston_FREEALL() \
    amdlibFree3DArrayComplex(tmpCpxVis);        

/**
 * Removes a constant phase (achromatic piston) on a complexVisibility vector.
 * Note this version artificially removes -1*pst since the pst sign in amdlib
 * is historically false.
 */
amdlibCOMPL_STAT amdlibCorrect3DVisTableFromAchromaticPiston(
                                            amdlibCOMPLEX ***cpxVisTable, 
                                            amdlibCOMPLEX ***cNopTable,
                                            int nbFrames, 
                                            int nbBases, 
                                            int nbLVis, 
                                            float *wlen,
                                            float **pst,
                                            amdlibERROR_MSG errMsg)
{

    int iFrame; 
    int iBase; 
    int lVis;
    double x;
    amdlibCOMPLEX phasor, cpxVis, cpxVisAvg;
    amdlibCOMPLEX *** tmpCpxVis = NULL;

    tmpCpxVis = amdlibAlloc3DArrayComplex(nbLVis, nbBases, nbFrames, errMsg);
    if (tmpCpxVis == NULL)
    {
        amdlibCorrect3DVisTableFromAchromaticPiston_FREEALL();
        return amdlibFAILURE;
    }

    for (iFrame = 0; iFrame < nbFrames; iFrame++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                x = (2 * M_PI/wlen[lVis]) * -1*pst[iFrame][iBase];
                phasor.re = cos(x);
                phasor.im = -sin(x);
                cpxVis.re = cpxVisTable[iFrame][iBase][lVis].re;
                cpxVis.im = cpxVisTable[iFrame][iBase][lVis].im;
                /* Store complex Product of the phasor with the coherent flux
                   in cNopTable*/
                amdlibCpxMul(cpxVis, phasor, cNopTable[iFrame][iBase][lVis]);
            }
        }
    }
    /* Recenter this by removing the mean. */
    /* Compute the mean over lambda */
    for (iFrame = 0; iFrame < nbFrames; iFrame++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            cpxVisAvg.re = 0.0;
            cpxVisAvg.im = 0.0;
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                cpxVisAvg.re += cNopTable[iFrame][iBase][lVis].re;
                cpxVisAvg.im += cNopTable[iFrame][iBase][lVis].im;
            }
            cpxVisAvg.re /= nbLVis;
            cpxVisAvg.im /= nbLVis;
            /* Store conjugate complex values in tmpCpxVis as (constant) 
             * vector */
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                tmpCpxVis[iFrame][iBase][lVis].re = cpxVisAvg.re;
                tmpCpxVis[iFrame][iBase][lVis].im = -cpxVisAvg.im;
            }
        }
    }
    /*Remove constant by multiplying the two */
    for (iFrame = 0; iFrame < nbFrames; iFrame++)
    {
        for (iBase = 0; iBase < nbBases; iBase++)
        {
            for (lVis = 0; lVis < nbLVis; lVis++)
            {
                phasor.re = tmpCpxVis[iFrame][iBase][lVis].re;
                phasor.im = tmpCpxVis[iFrame][iBase][lVis].im;
                cpxVis.re = cNopTable[iFrame][iBase][lVis].re;
                cpxVis.im = cNopTable[iFrame][iBase][lVis].im;
                /* Store Complex Product of the phasor with the coherent flux
                   in cNopTable*/
                amdlibCpxMul(cpxVis, phasor, cNopTable[iFrame][iBase][lVis]);
            }
        }
    }
    amdlibCorrect3DVisTableFromAchromaticPiston_FREEALL();

    return amdlibSUCCESS;
}

int amdlibFindIndexOfMinimum(double *data, int N)
{
    double min = data[0];
    int theIndex = 0;
    int i;
    for (i = 1; i < N; i++)
    {
        if (data[i] <= min)
        {
            min = data[i];
            theIndex = i;
        }
    }
    return (theIndex);
}
amdlibCOMPL_STAT amdlibComputeBaselines(amdlibISS_INFO *iss, int nbBases, 
                                        double duration)
{
    double ha,x,y,z,u,v,w;
    double xx[amdlibNBASELINE],yy[amdlibNBASELINE],zz[amdlibNBASELINE];
    double cl,sl,ch,sh,cd,sd;
    int i;

    /* play it for lst and lst+duration*/
    cl=cos(iss->geoLat);
    cd=cos(iss->dec);
    sl=sin(iss->geoLat);
    sd=sin(iss->dec);
    /* get baseline vectors in equatorial local coords */
    for (i = 0; i < nbBases; i++)
    {
        int tel1;
        int tel2;
        if (nbBases == 1)
        {
            tel1 = 0;
            tel2 = 1;
        }
        else
        {
            tel1 = amdlibMin((i % nbBases) , ((i+1) % nbBases) );
            tel2 = amdlibMax((i % nbBases) , ((i+1) % nbBases) );
        }

        /*axes x and y seem reverted wrt the aspro convention. I use aspro sign:*/
        x=(-1*iss->stationCoordinates[0][tel2])-(-1*iss->stationCoordinates[0][tel1]);
        y=(-1*iss->stationCoordinates[1][tel2])-(-1*iss->stationCoordinates[1][tel1]);
        z=iss->stationCoordinates[2][tel2]-iss->stationCoordinates[2][tel1];

        /* convert to equatorial local */
        xx[i]=0*x-sl*y+cl*z;
        yy[i]=x+0*y+0*z;
        zz[i]=0*x+cl*y+sl*z;
    }
    
    for (i = 0; i < nbBases; i++)
    {
        /* hour angle start */
        ha=fmod(iss->lst-iss->ra,2*M_PI);
        ch=cos(ha);
        sh=sin(ha);
        /* project onto star direction */
        u=sh*xx[i]+ch*yy[i];
        v=-sd*ch*xx[i]+sd*sh*yy[i]+cd*zz[i];
        w=cd*ch*xx[i]-cd*sh*yy[i]+sd*zz[i];
        /* Update PBL and PBLA at Start and End to enable further interpolation of
         * u and v in amdlibVisibilities. */
        iss->projectedBaseStart[i]=sqrt(u*u+v*v);
        iss->projectedAngleStart[i]=atan2(v,u)*180.0/M_PI;
        /* hour angle end. duration is in seconds of time, convert to radians */
        ha=fmod(iss->lst-iss->ra+(duration*M_PI/3600.0/12.0),2*M_PI);
        ch=cos(ha);
        sh=sin(ha);
        /* project onto star direction */
        u=sh*xx[i]+ch*yy[i];
        v=-sd*ch*xx[i]+sd*sh*yy[i]+cd*zz[i];
        w=cd*ch*xx[i]-cd*sh*yy[i]+sd*zz[i];
        /* Update PBL and PBLA at Start and End to enable further interpolation of
         * u and v in amdlibVisibilities. */
        iss->projectedBaseEnd[i]=sqrt(u*u+v*v);
        iss->projectedAngleEnd[i]=atan2(v,u)*180.0/M_PI;
    }
    return amdlibSUCCESS;
}

double amdlibAbacusErrPhi(double x)
/**
 * Estimate true phase rms from the cross-spectrum variance.
 * see Petrov, Roddier and Aime, JOSAA vol 3, N°5, may 1986 p 634.
 * and Petrov's Thesis, p. 50 ff.
 * I replace the piecewise interpolation usually used by a polynomial
 * approximation of the function:
 * if z=log10(y), 
 * then z=(C[1]*X^7+C[2]*X^6+C[3]*X^5+C[4]*X^4+C[5]*X^3+C[6]*X^2+C[7]*X+C[8])
 * and y=10^z.
 * where
 * C[01]= 2.71918080109099
 * C[02]=-17.1901043936273
 * C[03]= 45.0654103760899
 * C[04]=-63.4441678243197
 * C[05]= 52.3098941426378
 * C[06]=-25.8090699917488
 * C[07]= 7.84352873962491
 * C[08]=-1.57308595820081
 * This interpolation is valid in the range x=[0.1,1.74413]. 
 * Error is 1% everuwhere except above x=1.73 where it is 10%
 * Below x=0.1: y=x
 * Above x=M_PI/sqrt(3.0), y = blanking (impossible value)
 * Above x=1.74413, we take: y=0.691/(pi/sqrt(3.0)-x)
 */
{
    const double Asymptot=M_PI/sqrt(3.0);
    double c[8]={2.7191808010909,
                 -17.1901043936273,
                 45.0654103760899,
                 -63.4441678243197,
                 52.3098941426378,
                 -25.8090699917488,
                 7.84352873962491,
                 -1.57308595820081};
    
    double x2,x3,x4,x5,x6,x7,z;
    if(x>Asymptot)
    { 
        return((double)amdlibBLANKING_VALUE);
    }
    if(x>1.74413)
    {
        return(0.691/(Asymptot-x));
    }
    if(x<0.1)
    {
        return(x);
    }
    x2=x*x;
    x3=x2*x;
    x4=x2*x2;
    x5=x3*x2;
    x6=x3*x3;
    x7=x6*x;
    z=c[0]*x7+c[1]*x6+c[2]*x5+c[3]*x4+c[4]*x3+c[5]*x2+c[6]*x+c[7];
    return pow(10,z);
}
/*___oOo___*/
