/************************************************************************
* E.S.O. - VLT project
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* ambermgr  12/02/03  created
* mhein     25/03/03  amdlibGeneratePixelBiasData implemented
*/

/**
 * @file
 * Functions to handle pixel bias map.
 *
 * The pixel bias maps are needed in order to compensate for detector effects.
 * These maps are made by calculating an average image out of a sequence of raw
 * uncalibrated data. A pixel bias map is only validate for a specific subwindow
 * setup and exposure time. Normally a sequence of sky images are used to
 * generate a pixel bias map.  
 */

#define _POSIX_SOURCE 1

static char *rcsId="@(#) $Id: amdlibPixelBias.c,v 1.5 2011/01/24 12:39:17 agabasch Exp $"; 
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

/* 
 * System Headers
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "fitsio.h"

/* 
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"

/* 
 * Local function declaration 
 */
static void amdlibInitPixelBiasData(amdlibPIXEL_BIAS_DATA *pixelBias);
static amdlibCOMPL_STAT amdlibAllocatePixelBiasData
                                (amdlibRAW_DATA  *rawData,
                                 amdlibPIXEL_BIAS_DATA *pixelBias);
static void amdlibFreePixelBiasData(amdlibPIXEL_BIAS_DATA *pixelBias);

/* 
 * Public functions 
 */
/**
 * Release memory allocated to store pixel bias 
 *
 * @param pixelBias structure where pixel bias is stored
 */
void amdlibReleasePixelBiasData(amdlibPIXEL_BIAS_DATA *pixelBias)
{
    amdlibLogTrace("amdlibReleasePixelBiasData()");
    
    amdlibFreePixelBiasData(pixelBias);
    memset (pixelBias, '\0', sizeof(amdlibPIXEL_BIAS_DATA));    
}

/**
 * Set pixel bias map to the given value. 
 *
 * This function alloctes pixel bias structure corresponding to the raw data
 * structure (i.e. having same regions), and generate the pixel bias map by
 * computing the mean pixel value on all frames. 
 *
 * @param rawData structure where raw data corresponding to pixel bias is stored
 * @param pixelBias structure where pixel bias will be stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibGeneratePixelBiasData(amdlibRAW_DATA        *rawData,
                                             amdlibPIXEL_BIAS_DATA *pixelBias,
                                             amdlibERROR_MSG       errMsg)
{
    int     iRegion;
    int     iRow, iCol;
    int     iPixel;
    int     iFrame;
    int     regionSize;
    int     nbFrames;
    int     nbGoodFrames;
    double   sum;

    amdlibLogTrace("amdlibGeneratePixelBiasData()");
    
    /* Check raw data structure contains data */
    if (rawData->dataLoaded == amdlibFALSE)
    {
        amdlibSetErrMsg("The raw data structure does "
                        "not contain data. Check call to amdlibLoadRawData()");
        return amdlibFAILURE;
    }

    /* Check frame type */
    if (rawData->frameType != amdlibDARK_FRAME) 
    {
        amdlibLogWarning("Use of a non-dark frame type: proceed at your own "
                         "risks");
    }

    if (amdlibAllocatePixelBiasData(rawData, pixelBias) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for the pixel bias data");
        return amdlibFAILURE;
    }

    /* Calculate the bias image */
    for (iRow = 0; iRow < rawData->nbRows; iRow++)
    {
        for (iCol = 0; iCol < rawData->nbCols; iCol++)
        {
            iRegion      = iRow * rawData->nbCols + iCol;
            regionSize   = rawData->region[iRegion].dimAxis[0] *
                rawData->region[iRegion].dimAxis[1];
            nbFrames     = rawData->region[iRegion].dimAxis[2];
            for (iPixel = 0; iPixel < regionSize; iPixel++)
            {
                sum = 0.0;
                nbGoodFrames=0;
                for (iFrame = 0; iFrame < nbFrames; iFrame++)
                {
                    if ( rawData->timeTag[iFrame] != 0.0 )
                    {
                        nbGoodFrames++;
                        sum += rawData->region[iRegion].data[iFrame*regionSize 
                            + iPixel];
                    }
                }
                pixelBias->region[iRegion].data[iPixel] = sum / (float)nbGoodFrames;
            }
        }
    }

    return amdlibSUCCESS;
}

/**
 * Set pixel bias map to the given value. 
 *
 * This function allocates pixel bias structure corresponding to the raw data
 * structure (i.e. having same regions), and set all value of the pixel bias map
 * to the given value. 
 *
 * @param rawData structure where raw data corresponding to pixel bias is stored
 * @param pixelBias structure where pixel bias will be stored
 * @param value pixel bias value.
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibSetPixelBiasData (amdlibRAW_DATA        *rawData,
                                         amdlibPIXEL_BIAS_DATA *pixelBias,
                                         float                 value,
                                         amdlibERROR_MSG       errMsg)
{
    int     iRegion;
    int     iRow, iCol;
    int     iPixel;
    int     regionSize;

    amdlibLogTrace("amdlibSetPixelBiasData()");
    
    /* Check raw data structure contains data */
    if (rawData->dataLoaded == amdlibFALSE)
    {
        amdlibSetErrMsg("The raw data structure does "
                        "not contain data. Check call to amdlibLoadRawData()");
        return amdlibFAILURE;
    }

    /* Allocate memory for data */
    if (amdlibAllocatePixelBiasData(rawData, pixelBias) != amdlibSUCCESS)
    {
        amdlibSetErrMsg("Could not allocate memory for the pixel bias data");
        return amdlibFAILURE;
    }

    /* Set the bias image */
    for (iRow = 0; iRow < rawData->nbRows; iRow++)
    {
        for (iCol = 0; iCol < rawData->nbCols; iCol++)
        {
            iRegion      = iRow * rawData->nbCols + iCol;
            regionSize   = rawData->region[iRegion].dimAxis[0] *
                rawData->region[iRegion].dimAxis[1];
            for (iPixel = 0; iPixel < regionSize; iPixel++)
            {
                pixelBias->region[iRegion].data[iPixel] = value;
            }
        }
    }

    return amdlibSUCCESS;
}

/*
 * Local functions
 */
/**
 * Initialize pixel bias structure.
 *
 * @param pixelBias pointer to pixel bias structure
 */
void amdlibInitPixelBiasData(amdlibPIXEL_BIAS_DATA *pixelBias)
{
    amdlibLogTrace("amdlibInitPixelBiasData()");
    
    /* Initialize data structure */
    memset (pixelBias, '\0', sizeof(amdlibPIXEL_BIAS_DATA));
    pixelBias->thisPtr = pixelBias;
}

/**
 * Allocate memory for pixel bias structure.
 *
 * This function allocates memory for pixel bias which will result from
 * computation of the given raw data. 
 * 
 * @param rawData structure where raw data which will be used to compute pixel
 * bias is stored
 * @param pixelBias structure where pixel bias will be stored
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibAllocatePixelBiasData(amdlibRAW_DATA        *rawData,
                                             amdlibPIXEL_BIAS_DATA *pixelBias)
{
    int  iRegion;

    amdlibLogTrace("amdlibAllocatePixelBiasData()");
    
    if (rawData->thisPtr != rawData)
    {
        return amdlibFAILURE;
    }

    /* If pixel bias data structure is not initialized, do it */
    if (pixelBias->thisPtr != pixelBias)
    {
        amdlibInitPixelBiasData(pixelBias);
    }

    amdlibFreePixelBiasData(pixelBias);
    pixelBias->expTime = rawData->expTime;
    pixelBias->nbRows = rawData->nbRows;
    pixelBias->nbCols = rawData->nbCols;
    pixelBias->imagingDetector = rawData->imagingDetector;
    pixelBias->nbRegions = rawData->nbRegions;

    /* Allocate memory for storing information of regions */
    if (amdlibAllocateRegions(&pixelBias->region, 
                              pixelBias->nbRegions) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }

    /* Allocate memory for storing region data */
    for (iRegion = 0; iRegion < pixelBias->nbRegions; iRegion++)
    {
        int nbOfElements;
        
        memcpy(&pixelBias->region[iRegion],
               &rawData->region[iRegion],sizeof(amdlibREGION));
        pixelBias->region[iRegion].dimAxis[2] = 1;
        pixelBias->region[iRegion].dmp = NULL;
        pixelBias->region[iRegion].dmc = NULL;
        nbOfElements = rawData->region[iRegion].dimAxis[0] *
            rawData->region[iRegion].dimAxis[1];
        
        /* Allocate memory for data */
        pixelBias->region[iRegion].data = calloc(sizeof(float), nbOfElements);
        if (pixelBias->region[iRegion].data == NULL)
        {
            return amdlibFAILURE;
        }
    }
    return amdlibSUCCESS;
}

/**
 * Free memory allocated memory of pixel bias structure.
 *
 * This function frees previously allocated memory (if any) where pixel bias has
 * been stored. 
 * 
 * @param pixelBias structure where pixel bias is stored
 */
void amdlibFreePixelBiasData(amdlibPIXEL_BIAS_DATA *pixelBias)
{
    amdlibLogTrace("amdlibFreePixelBiasData()");
    
    if (pixelBias->thisPtr != pixelBias)
    {
        amdlibInitPixelBiasData(pixelBias);
        return;
    }

    amdlibFreeRegions(&pixelBias->region, pixelBias->nbRegions);
}

/*___oOo___*/
