/************************************************************************
 * E.S.O. - VLT project_
 *
 * who       when      what
 * --------  --------  ----------------------------------------------
 * gduvert   23/08/05  corrected severe memory loss due to lazy positioning
 *                     of amdlibShift_FREEALL.
 * gduvert   22/08/05  replaced DEBUG with VERBOSE to avoid spurious outputs
 *                     from these routines when called from realtime programs
 *                     at ESO that are compiled with DEBUG defined. 
 * mkiekebu 2005-0708  added more control over memory allocation.
 * fmillour  08/07/05  replaced amdlibComputeShift with new algoritmh,
 *                     normally more robust wrt large shifts. Old algo
 *                     is maintained in amdlibComputeShiftByFTSlope.
 *                     gd note: both seem to give same results!
 * gduvert   22/04/05  corrected both amdlibShift & amdlibComputeShift
 *                     code.
 * fmillour  24/08/03  Routines to get/set a non integer displacement
 *                     of pixels in an array.
 */

/**
 * @file
 * Functions related to floating array shift. 
 */
#define _POSIX_SOURCE 1

static char *rcsId="@(#) $Id: amdlibShift.c,v 1.5 2011/01/24 12:43:59 agabasch Exp $";
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

/*
 * Macros definitions
 */

/*
 * System Headers
 */
#include <string.h>
#include <math.h>

/*
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"

/*
 * Protected function
 */
/** Useful macro to free all allocated memory */
#define amdlibShift_FREEALL()  amdlibFree(drc);             \
                               amdlibFree(tabInPeriodized); \
                               amdlibFree(fftDirect);       \
                               amdlibFree(phasor);          \
                               amdlibFree(phasedFft);       \
                               amdlibFree(tabDecalPeriodized);
/**
 * Shift array
 *
 * This function shifts array tabIn by 'shift' (possibly non-integer) number in
 * tabOut, wrapping around values.
 *
 * @param nbElem number of element in arrays
 * @param tabIn input array 
 * @param shift shift to perform 
 * @param tabOut resulting shifted array 
 * @param errMsg error description message returned if function fails.
 * 
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibShift(/* INPUT */
                             int             nbElem,
                             fftw_real       *tabIn,
                             fftw_real       shift,
                             /* OUTPUT */
                             fftw_real       *tabOut,
                             amdlibERROR_MSG errMsg)
{
    int i;

    /*  Get the integer part (ip) and fractionnal part (fp)
     *  of the offset to apply to tabIn */
    int   ipShift = (int)shift;
    fftw_real fpShift = shift - ipShift;
 
    /*  Get the values of the first and last points to substract
     *  the affine part of the signal */
    fftw_real m1 = tabIn[0];
    fftw_real m2 = tabIn[nbElem - 1];
    fftw_real phi;
    fftw_real dr;
    rfftw_plan p;

    /* Arrays */
    fftw_real *drc = NULL;
    fftw_real *tabInPeriodized = NULL;
    fftw_real *fftDirect = NULL;
    fftw_real *phasor = NULL;
    fftw_real *phasedFft = NULL;
    fftw_real *tabDecalPeriodized = NULL;

    amdlibLogTrace("amdlibShift()");

#if TEST
    amdlibLogTest("amdlibShift()");
    amdlibLogTest("    Array size = %d", nbElem);
    amdlibLogTest("    Shift      = %.3f", shift);
#endif  

    /* Check shift value; could not be greater then array size */
    if (abs(ipShift) > nbElem)
    {
        amdlibSetErrMsg("Invalid shift %.3f; greater than array size %d",
                        shift, nbElem);	
        return amdlibFAILURE;
    }

    /* Define work arrays */

    drc = calloc(nbElem, sizeof(fftw_real));
    if (drc == NULL)
    {
        amdlibShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (drc)");
        return amdlibFAILURE;
    }
    
    tabInPeriodized = calloc(nbElem, sizeof(fftw_real));
    if (tabInPeriodized == NULL)
    {
        amdlibShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (tabInPeriodized)");
        return amdlibFAILURE;
    }
    
    fftDirect = calloc(nbElem, sizeof(fftw_real));
    if (fftDirect == NULL)
    {
        amdlibShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (fftDirect)");
        return amdlibFAILURE;
    } 

    phasor = calloc(nbElem, sizeof(fftw_real));
    if (phasor == NULL)
    {
        amdlibShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (phasor)");
        return amdlibFAILURE;
    }  
    
    phasedFft = calloc(nbElem, sizeof(fftw_real));
    if (phasedFft == NULL)
    {
        amdlibShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (phasedFft)");
        return amdlibFAILURE;
    }  
    
    tabDecalPeriodized = calloc(nbElem, sizeof(fftw_real));
    if (tabDecalPeriodized == NULL)
    {
        amdlibShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (tabDecalPeriodized)");
        return amdlibFAILURE;
    }  

    /* */
    if (fpShift != 0.0)
    {
        for (i=0;i<nbElem;i++)
        {
            /*  The value to substract to the signal before the shift */
            dr = ((m2-m1) / (nbElem - 1)) * i + m1;

            /*  The value to add again to the periodized signal after the 
             *  shift */
            drc[i] = dr + ((m2 - m1) / (nbElem - 1)) * -fpShift;
            tabInPeriodized[i] = tabIn[i] - dr;
        }  

#if TEST
        amdlibLogTest("drc =[");
        for (i = 0; i < nbElem - 1; i++)
        {
            amdlibLogTest("%f, ", drc[i]);
        }
        amdlibLogTest("%f]", drc[nbElem - 1]);

        amdlibLogTest("tabInPeriodized =[");
        for (i = 0; i < nbElem - 1; i++)
        {
            amdlibLogTest("%f, ", tabInPeriodized[i]);
        }
        amdlibLogTest("%f]", tabInPeriodized[nbElem-1]);
#endif

#if TEST
        amdlibLogTest("phi =[");
#endif

        phi = 0; /*(2. *  M_PI * fpShift / nbElem) * 1.0;*/
        phasor[0] =  cos(phi);  /* DC component */

        for (i = 1; i < (nbElem + 1) / 2; i++)
        {
            /*  The phasor to apply to the fft of the periodized signal
             *  to shift it of fpShift */
            phi = (2. *  M_PI * fpShift / nbElem) * (i);
#if TEST
            amdlibLogTest("%f, ", phi);
#endif

            phasor[i]         =  cos(phi);
            phasor[nbElem - i] =  sin(phi);
        }
        if (nbElem % 2 == 0) /* nbElem is even */
        {
            phi = M_PI * fpShift;
            phasor[nbElem / 2] =  cos(phi);  /* Nyquist freq. */
        }

#if TEST
        amdlibLogTest("]");

        amdlibLogTest("phasor =[");
        for (i = 0; i < nbElem - 1; i++)
        {
            amdlibLogTest("%f, ", phasor[i]);
        }
        amdlibLogTest("%f]", phasor[nbElem - 1]);
#endif

        /*  The shift itself */
        p = rfftw_create_plan(nbElem, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
        rfftw_one(p, tabInPeriodized, fftDirect);
        rfftw_destroy_plan(p);

        phasedFft[0] = phasor[0] * fftDirect[0];
        for (i = 1; i < (nbElem + 1) / 2; i++)
        { 
            /* f(x-x0)<=>F(u)exp(i*2*PI/N*u*x0) */
            /* Real part */
            phasedFft[i] = phasor[i] * fftDirect[i] +
                phasor[nbElem - i] * fftDirect[nbElem - i];

            /* Imaginary part */
            phasedFft[nbElem-i] = phasor[i] * fftDirect[nbElem - i] -
                phasor[nbElem - i] * fftDirect[i];
        }
        
        if (nbElem % 2 == 0) /* nbElem is even Real part only */
        {
            phasedFft[nbElem / 2] = 2 * phasor[nbElem / 2] * 
                fftDirect[nbElem / 2];
        }

#if TEST
        amdlibLogTest("fftDirect =[");
        for (i = 0; i < nbElem - 1; i++)
        {
            amdlibLogTest("%f, ", fftDirect[i]);
        }
        amdlibLogTest("%f]", fftDirect[nbElem-1]);

        amdlibLogTest("phasedFft =[");
        for (i = 0; i < nbElem - 1; i++)
        {
            amdlibLogTest("%f, ", phasedFft[i]);
        }
        amdlibLogTest("%f]", phasedFft[nbElem-1]);
#endif

        p = rfftw_create_plan(nbElem, FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE);
        rfftw_one(p, phasedFft, tabDecalPeriodized);
        rfftw_destroy_plan(p);

        for (i = 0; i < nbElem; i++)
        {
            tabIn[i] = (1.0 / nbElem) * tabDecalPeriodized[i] + 
                drc[i];
        }
    }
    else if (shift == 0.0) 
    {
        memcpy(tabOut, tabIn, (size_t)nbElem * sizeof(fftw_real));
        amdlibShift_FREEALL();	
        return amdlibSUCCESS;
    }

    if (ipShift < 0)
    {
        for (i = 0; i < -ipShift; i++)
        {
            tabOut[i + ipShift + nbElem] = tabIn[i];
        }
        for (i = -ipShift; i<nbElem; i++)
        {
            tabOut[i + ipShift] = tabIn[i];
        }
    }
    else
    {
        for (i = 0; i < nbElem; i++)
        {
            tabOut[i] = (i < ipShift) ? 
                tabIn[i - ipShift + nbElem]:tabIn[i - ipShift];
        }
    }


#if TEST
    amdlibLogTest("tabDecalPeriodized =[");
    for (i = 0; i < nbElem - 1; i++)
    {
        amdlibLogTest("%f, ", tabDecalPeriodized[i]);
    }
    amdlibLogTest("%f]", tabDecalPeriodized[nbElem - 1]);

    amdlibLogTest("tabOut =[");
    for (i = 0; i < nbElem - 1; i++)
    {
        amdlibLogTest("%f, ", tabOut[i]);
    }
    amdlibLogTest("%f]", tabOut[nbElem - 1]);

    amdlibLogTest("tabIn =[");
    for (i = 0; i < nbElem - 1; i++)
    {
        amdlibLogTest("%f, ", tabIn[i]);
    }
    amdlibLogTest("%f]", tabIn[nbElem - 1]);
    amdlibLogTest("tabOut (entier) =[");
    for (i = 0; i < nbElem - 1; i++)
    {
        amdlibLogTest("%f, ", tabOut[i]);
    }
    amdlibLogTest("%f]", tabOut[nbElem - 1]);

    if (fpShift == 0.0) 
    {
        amdlibLogTest("tabOut (entier decal) =[");
        for (i = 0; i < nbElem - 1; i++)
        {
            amdlibLogTest("%f, ", tabOut[i]);
        }
        amdlibLogTest("%f]", tabOut[nbElem - 1]);
    }
#endif      

    amdlibShift_FREEALL();

    return amdlibSUCCESS;
}

/** Useful macro to free all allocated memory */
#define amdlibComputeShift_FREEALL() amdlibFree(fft_tab2);      \
                                     amdlibFree(fft_tab1);      \
                                     amdlibFree(interSpectrum); \
                                     amdlibFree(convolPadded);

/**
 * Computes shift between to value arrays
 *
 * This function computes the shift (possibly non-integer) between two arrays.
 *
 * @param nbElem number of element in arrays
 * @param tab1 first array 
 * @param tab2 second array 
 * @param shift computed shift between arrays
 * @param sigma2Shift error on computed shift
 * @param errMsg error description message returned if function fails.
 * 
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibComputeShift(/* INPUT */
                                    int             nbElem,
                                    fftw_real       *tab1,
                                    fftw_real       *tab2,
                                    /* OUTPUT */
                                    fftw_real       *shift,
                                    fftw_real       *sigma2Shift,
                                    amdlibERROR_MSG errMsg)
{
    int i;
    int nPhi = (nbElem + 1) / 2;

    /* Interpolated spectrum */
    fftw_real *interSpectrum = NULL;
    fftw_real *convolPadded = NULL;
    fftw_real *fft_tab1 = NULL;
    fftw_real *fft_tab2 = NULL ;
    int padFactor;
    int mxx;
    double max;


    rfftw_plan p1;
    rfftw_plan p2;

    amdlibLogTrace("amdlibComputeShift()");
    
    padFactor = 32;

    /* Allocate memory */
    interSpectrum = calloc(nbElem * padFactor, sizeof(fftw_real));
    if (interSpectrum == NULL)
    {
        amdlibComputeShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (interSpectrum)");		
        return amdlibFAILURE;
    }
    convolPadded = calloc(nbElem * padFactor, sizeof(fftw_real));
    if (convolPadded == NULL)
    {
        amdlibComputeShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (convolPadded)");
        return amdlibFAILURE;
    }
    fft_tab1     = calloc(nbElem, sizeof(fftw_real));
    if (fft_tab1 == NULL)
    {
        amdlibComputeShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (fft_tab1)");
        return amdlibFAILURE;
    }
    fft_tab2     = calloc(nbElem, sizeof(fftw_real));
    if (fft_tab2 == NULL)
    {
        amdlibComputeShift_FREEALL();
        amdlibSetErrMsg("Could not allocate memory (fft_tab2)");
        return amdlibFAILURE;
    }

    p1 = rfftw_create_plan(nbElem, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE);
    rfftw_one(p1, tab1, fft_tab1);
    rfftw_one(p1, tab2, fft_tab2);

    interSpectrum[0] = fft_tab1[0] * fft_tab2[0];
    for (i = 1; i < nPhi; i++)
    {
        interSpectrum[i] =
            fft_tab1[i] * fft_tab2[i] +
            fft_tab1[nbElem - i] * fft_tab2[nbElem - i];

        interSpectrum[nbElem * padFactor - i] =
            fft_tab1[nbElem - i] * fft_tab2[i] -
            fft_tab1[i] * fft_tab2[nbElem - i];
    }
    if (nbElem % 2 == 0) /* nbElem is even */
    {
        interSpectrum[nPhi - 1] = fft_tab1[nbElem / 2] * fft_tab2[nbElem / 2] +
            fft_tab1[nbElem / 2] * fft_tab2[nbElem / 2];
    }

    p2 = rfftw_create_plan(nbElem * padFactor, FFTW_COMPLEX_TO_REAL, 
                           FFTW_ESTIMATE);

    rfftw_one(p2, interSpectrum, convolPadded);

    mxx = 0;
    max = convolPadded[mxx];
    for (i = 1; i < nbElem * padFactor; i++)
    {
        if(convolPadded[i] > max)
        {
            max = convolPadded[i];
            mxx = i;
        }
    }

    if (mxx > nbElem * padFactor / 2)
    {
        mxx = mxx - nbElem * padFactor;
    }

    *shift = -1 * (fftw_real) mxx / (fftw_real) padFactor;
    *sigma2Shift = 1.0 / (fftw_real) padFactor;

    rfftw_destroy_plan(p1);
    rfftw_destroy_plan(p2);

    /* Free memory */
    amdlibComputeShift_FREEALL();

    return amdlibSUCCESS;
}


