/*******************************************************************************
 * E.S.O. - VLT project
 *
 * who       when      what
 * --------  --------  ----------------------------------------------
 * gduvert   12/06/07  Modified to permit a better spectral calibration
 *                     directly at the amdlibComputeP2vm level.
 * gduvert   05/02/06  - removed all computations for High Spectral resolution 
 *                       Data (hardware prevents it)
 * gzins     26/01/06  - added check of the computed shift
 * gduvert   22/08/05  - removed spurious reference to DEBUG, not used here.
 * gduvert   08/07/05  - updated JHK bands limits to avoid at 5% level inter-
 *                       band regions where in JHK mode two fringe systems from
 *                       adjacent bands are superimposed.
 * gduvert   10/06/05  - removed computation of shift between spectra in the
 *                       interferometric channel from one beam to another since
 *                       by construction the spectra are not shifted (otherwise there
 *                       would not be any fringes at all). Furthermore, the check was
 *                       wrong because the spectra are different from beam to beam and
 *                       not directly comparable.
 *                     - use the setup of shutters to check what bands are used 
 *                       (mainly JHK mode) to find the correct spectral calibration spectrum
 *                       to be used as a template when aligning observed spectrum 
 *                       vs. reference spectrum.
 *                     - removed amdlibCheckJHKModeSpectralDispersion and its use.
 * gduvert   30/05/05  - moved amdlibGetBand here for general use.
 *                     - renamed CheckJHKModeSpectralDispersion in 
 *                       "amdlibCheckJHKModeSpectralDispersion"
 * gduvert   22/04/05  - comparison of photometric wrt interferometric
 *                       channel with amdlibComputeShift is done frame
 *                       by frame, not wrt a 'global' interferometric channel
 *                       that was the sum of the 2 or 3 ICs. For some reason,
 *                       the ICs seem at time not to superimpose well, which
 *                       is totally out of specs.
 * gduvert   26/10/04  - add CheckJHKModeSpectralDispersion function
 *                     - modify the retrieve of correct photometry and the 
 *                       compute shift wrt. interferometric channel
 *                     - test shift of interf.chan of frame 2 and 3  wrt. 
 *                       interf. chan of frame 1 (stored in workSpace1)
 * gduvert   07/10/04  created
 */

/**
 * @file
 * Functions to perform spectral calibration
 */

#define _POSIX_SOURCE 1

static char *rcsId="@(#) $Id: amdlibSpectralCalibration.c,v 1.5 2011/01/24 12:44:00 agabasch Exp $";
static void *use_rcsId = (0 ? &use_rcsId : (void*)&rcsId);

/*
 * System Headers
 */
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "fitsio.h"

/*
 * Local Headers
 */
#include "amdlib.h"
#include "amdlibProtected.h"

/* 
 * Public function 
 */

/**
 * Perform spectral calibration.
 *
 * This function perform the spectral calibration. This consists in :
 *      \li computing the photometric channel shift wrt. interferometric channel
 *      \li calibration spectral dispersion table (in low JHK mode only)
 *
 * @param scData data used for spectral calibration.
 * @param scType spectral calibration type (amdlibSC_2T or amdlibSC_3T).
 * @param waveData structure where calibration wave data will be stored
 * @param errMsg error description message returned if function fails.
 *
 * @return
 * amdlibSUCCESS on successful completion. Otherwise amdlibFAILURE is returned.
 */
amdlibCOMPL_STAT amdlibPerformSpectralCalibration(amdlibSC_INPUT_DATA *scData,
                                                  amdlibSC_TYPE       scType,
                                                  amdlibWAVEDATA      *waveData,
                                                  amdlibERROR_MSG     errMsg)
{
    int nbTel;
    int beam;
    int nbChannels;
    float *spectrum;
    int hasAssociatedP2VMFlux[3]={0,0,0};
    fftw_real intfSpec [amdlibNB_SPECTRAL_CHANNELS];
    fftw_real photoSpec[amdlibNB_SPECTRAL_CHANNELS];

    char keywValue[amdlibKEYW_VAL_LEN+1];    
 
    amdlibLogTrace("amdlibPerformSpectralCalibration()"); 
    
    /* Init the number of telescopes and baselines according to SC type */
    if (scType == amdlibSC_2T)
    {
        nbTel = 2;
    }
    else if (scType == amdlibSC_3T)
    {
        nbTel = 3;
    }
    else
    {
        amdlibSetErrMsg("Invalid type %d (see amdlibSC_TYPE)", (int)scType);
        return amdlibFAILURE;
    }
    
    /* Get instrument configuration */
    memset(keywValue, '\0', (amdlibKEYW_VAL_LEN+1)*sizeof(char));
    if (amdlibGetInsCfgKeyword(&(scData->rawData[0].insCfg), 
                               "HIERARCH ESO OCS OBS SPECCONF",
                               keywValue, errMsg) != amdlibSUCCESS)
    {
        return amdlibFAILURE;
    }

    /* If instrument configuration is High Resolution, Do not even attempt
     * to compute offsets, hardware currently implement prevents it. Just
     * pretend that the offsets have been calibrated and equal 0*/
    if (strstr(keywValue, "High_") != NULL)
    {
        for (beam = 0; beam < nbTel; beam++)
        {
        waveData->photoOffset[beam] = 0.0;
        amdlibLogInfo("High-resolution data");
        amdlibLogInfoDetail("Photometric channel %d set to offset 0", beam);
        }
        return amdlibSUCCESS;
    }
    
    /* Reset error message */
    memset(errMsg, '\0', sizeof(amdlibERROR_MSG));
 
    /* Check the science data have been stored and calibrated */
    for (beam = 0; beam < nbTel; beam++)
    {
        int frameType;

        /* Set frame type (see amdlibFRAME_TYPE) */
        frameType = beam + amdlibTEL1_FRAME; 

        /* Check the corresponding science data has been loaded */
        if (scData->dataLoaded[beam] == amdlibFALSE)
        {
            amdlibSetErrMsg("Missing data file corresponding to the beam %d "
                            "(FRAMETYPE %d, see amdlibFRAME_TYPE)", 
                            beam, frameType);
            return amdlibFAILURE;
        }
        /* Check if data set also has flux correction table (e.g., from P2VM)*/
        if (scData->dataLoaded[beam+3] == amdlibTRUE)
        {
            hasAssociatedP2VMFlux[beam]=1;
        }
    }

    /* Retrieve the number of spctral channels */
    nbChannels = scData->rawData[0].region[0].dimAxis[1];

    /* For each telescope (i.e. beam), retrieve correct photometry and compute
     * shift wrt. interferometric channel */
    for (beam = 0; beam < nbTel; beam++)
    {
        fftw_real shift, sigma2Shift;
        int l;

        /* Allocate memory for spectrum */
        spectrum = (float *)calloc(nbChannels, sizeof(float));
        if (spectrum == NULL)
        {
            amdlibSetErrMsg("Could not allocate memory for spectrum");
            return amdlibFAILURE;
        }
        
        /* Compute spectrum for interferometric beam */
        if (amdlibSumAndPackData(&(scData->rawData[beam]), 
                                 amdlibTRUE, amdlibFALSE, amdlibTRUE,
                                 3, &spectrum, errMsg) != amdlibSUCCESS)
        {
            free(spectrum);
            return amdlibFAILURE;
        }

        /* Store interferometric spectrum into 512-element array for 'fft' */
        memset(intfSpec, '\0', amdlibNB_SPECTRAL_CHANNELS*sizeof(*intfSpec));
        for (l = 0; l < nbChannels; l++)   
        {
            intfSpec[l] = spectrum[l];
        }
          
        /* if associated flux calibration frame is present, and we are MR (HR?) correct it*/
        if (hasAssociatedP2VMFlux[beam]==1 && (strstr(keywValue, "Low_JHK ") == NULL))
        {
            if (amdlibSumAndPackData(&(scData->rawData[beam+3]), 
                                     amdlibTRUE, amdlibFALSE, amdlibTRUE,
                                     3, &spectrum, errMsg) != amdlibSUCCESS)
            {
                free(spectrum);
                return amdlibFAILURE;
            }
            /*normalise intfSpec with this spectrum*/
            for (l = 0; l < nbChannels; l++)   
            {
                intfSpec[l] /= spectrum[l];
            }
        }
        
        /* Compute spectrum for photometric beam */
        if (amdlibSumAndPackData(&(scData->rawData[beam]), 
                                 amdlibTRUE, amdlibFALSE, amdlibTRUE,
                                 beam, &spectrum, errMsg) != amdlibSUCCESS)
        {
            return amdlibFAILURE;
        }

        /* Store photometric spectrum into 512-element array for 'fft' */
        memset(photoSpec, '\0', amdlibNB_SPECTRAL_CHANNELS*sizeof(*photoSpec));
        for (l = 0; l < nbChannels; l++)   
        {
            photoSpec[l] = spectrum[l];
        }

        /* if associated flux calibration frame is present, correct it*/
        if (hasAssociatedP2VMFlux[beam]==1 && (strstr(keywValue, "Low_JHK ") == NULL) )
        {
            if (amdlibSumAndPackData(&(scData->rawData[beam+3]), 
                                     amdlibTRUE, amdlibFALSE, amdlibTRUE,
                                     beam, &spectrum, errMsg) != amdlibSUCCESS)
            {
                return amdlibFAILURE;
            }
            /*normalise photoSpec with this spectrum*/
            for (l = 0; l < nbChannels; l++)   
            {
                photoSpec[l] /= spectrum[l];
            }
        }

        /* Compute shift between the interferometric and photometric channels */
        if (amdlibComputeShift(nbChannels, intfSpec, photoSpec, 
                              &shift, &sigma2Shift, errMsg) == amdlibFAILURE)
        {	    
            free(spectrum);
            return amdlibFAILURE;
        }

        /* Check shift value; must be less than 5 pixels */
        if (fabs(shift) > 5.0)
        {
            free(spectrum);
            amdlibSetErrMsg("Shift (%.2f) between photometric channel #%d "
                            "and interferometric channel too high",
                            shift, beam + 1);
            return amdlibFAILURE;
        }
        
        /* Store shift */
        waveData->photoOffset[beam] = shift;
        amdlibLogInfoDetail("Photometric channel %d has offset %f (%f)", 
                            beam, shift, sigma2Shift);
        
        free(spectrum);
    }


    /* If it is low JHK mode, calibrate spectral dispersion table using image of
     * the beam 1 */
    if (strstr(keywValue, "Low_JHK ") != NULL)
    {
        int shutterNum;
        int i, l;
        int startPixel;

        /* Assuming all bands are open */
        int shutterPos[amdlibNB_BANDS] = {1, 1, 1};
        amdlibINS_CFG *insCfg = &scData->rawData[1].insCfg;

        /* Reference spectrum */
        fftw_real refSpec[amdlibNB_SPECTRAL_CHANNELS];
        fftw_real max;
        fftw_real shift, sigma2Shift;

        /* Scan all configuration keywords  */
        for (i=0; i < insCfg->nbKeywords; i++)
        {
            /* Look for SFK shutters;  do not check the specific BSL shutters */
            if ((strstr(insCfg->keywords[i].name,
                        "HIERARCH ESO INS SHUT") != NULL) &&
                (strstr(insCfg->keywords[i].name, " ST") != NULL))
            {

                /* Get the shutter number; used to know the optical path to
                 * which it belongs */
                sscanf(insCfg->keywords[i].name, 
                        "HIERARCH ESO INS SHUT%d", &shutterNum);   
                /* If shutter is closed */
                if (strstr(insCfg->keywords[i].value, "F") != 0)
                {
                    /* Set correspondind band to close */
                    switch(shutterNum)
                    {
                        /* Shutters of J band */
                        case 10:
                        case 11:
                            shutterPos[amdlibJ_BAND] = 0;
                            break;
                        case 12:
                            if (nbTel == 3)
                            {
                                shutterPos[amdlibJ_BAND] = 0;
                            }
                            break;
                            /* Shutters of H band */
                        case 7:
                        case 8:
                            shutterPos[amdlibH_BAND] = 0;
                            break;
                        case 9:
                            if (nbTel == 3)
                            {
                                shutterPos[amdlibH_BAND] = 0;
                            }
                            break;
                            /* Shutters of K band */
                        case 4:
                        case 5:
                            shutterPos[amdlibK_BAND] = 0;
                            break;
                        case 6:
                            if (nbTel == 3)
                            {
                                shutterPos[amdlibK_BAND] = 0;
                            }
                            break;
                    }
                }
            }
        }
        /* If all bands are off, return an error */
        if ((shutterPos[amdlibJ_BAND] == 0) &&
            (shutterPos[amdlibH_BAND] == 0) &&
            (shutterPos[amdlibK_BAND] == 0))
        {
            amdlibSetErrMsg("Unable to build JHK reference Spectrum: check "
                            "shutter status !");
            return amdlibFAILURE;
        }

        /* Get reference spectrum corresponding to the instrument
         * configuration */
        if (amdlibGetRefLowJHKSpectrumForCal(shutterPos, refSpec) 
                                            == amdlibFAILURE)
        {
            return amdlibFAILURE;
        }
        
        /* Allocate memory for spectrum */
        spectrum = (float *)calloc(nbChannels, sizeof(float));
        
        /* Compute spectrum for interferometric beam */
        if (amdlibSumAndPackData(&(scData->rawData[0]), 
                                 amdlibTRUE, amdlibFALSE, amdlibTRUE,
                                 3, &spectrum, errMsg) != amdlibSUCCESS)
        {
            free(spectrum);
            return amdlibFAILURE;
        }

        /* Get the start pixel; -1 because this value is on the detector where
         * pixels numbers start at 1 */
        startPixel = scData->rawData[0].region[3].corner[1] - 1;
        /* Store interferometric spectrum into 512-element array for 'fft'
         * taking into account the start pixel. */
        memset(intfSpec, '\0', amdlibNB_SPECTRAL_CHANNELS*sizeof(*intfSpec));
        for (l = 0; l < nbChannels; l++)   
        {
            intfSpec[startPixel + l] = spectrum[l];
         }
        free(spectrum); 

        /* Since spectra have now different powers, normalize the spectra
         * observed to unity */
        max = intfSpec[0];
        for (l = 0; l < amdlibNB_SPECTRAL_CHANNELS; l++)
        {
            if (intfSpec[l] > max)
            {
                max = intfSpec[l];
            }
        }
        for (l = 0; l < amdlibNB_SPECTRAL_CHANNELS; l++)   
        {
            intfSpec[l] /= max;
        }

        /* Compute shift */
        if (amdlibComputeShift(amdlibNB_SPECTRAL_CHANNELS, refSpec,
                               intfSpec, &shift, &sigma2Shift,
                               errMsg) == amdlibFAILURE)
        {
            return amdlibFAILURE;
        }
        amdlibLogInfo("Wavelength table has offset %f (%f)", 
                      shift, sigma2Shift);

        /* Check offset value; must be less than 10 pixels */
        if (fabs(shift) > 10.0)
        {
            free(spectrum);
            amdlibSetErrMsg("Offset (%.2f) of wavelength table too high",
                            shift);
            return amdlibFAILURE;
        }

        /* Update wavelength and bandwidth by removing data, inserting reference
         * spectrum wavelength at the good location and shifting this by the
         * nint of measured shift value (this to maintain zeroes in all
         * wavelength undefined -- with the non-integer shift method, invalid
         * wavelength would contain noise, not zeroes)
         */
        /* Get reference spectral dispersion table for low JHK mode */
        if (amdlibGetRefLowJHKSpectralDispersion(refSpec) == amdlibFAILURE)
        {
            return amdlibFAILURE;
        }
        /* Shift reference spectral dispersion table of the measured shift */
        if (amdlibShift(amdlibNB_SPECTRAL_CHANNELS, refSpec, shift,
                        waveData->wlen, errMsg) == amdlibFAILURE)
        {
            return amdlibFAILURE;
        }

        /* Set all wavelengths that do not belong to J, H or K band to 0 */
        l = 0;
        while (l < amdlibNB_SPECTRAL_CHANNELS)
        {
            if (amdlibGetBand(waveData->wlen[l]) == amdlibUNKNOWN_BAND)
            {
                waveData->wlen[l] = 0.;
                l++;
            }
            else 
            {
                break;
            }
        }
        i = amdlibNB_SPECTRAL_CHANNELS - 1;
        while (i > l)
        {
            if (amdlibGetBand(waveData->wlen[i]) == amdlibUNKNOWN_BAND)
            {
                waveData->wlen[i] = 0.;
                i--;
            }
            else 
            {
                break;
            }
        }

        /* Compute bandwidth */
        if (amdlibComputeBandwidth(waveData) == amdlibFAILURE)
        {
            amdlibSetErrMsg("Could not compute bandwidth");
            return amdlibFAILURE;
        }
    }

    return amdlibSUCCESS;
}

/* Protected functions */
/**
 * Compute bandwidth
 *
 * This function compute the bandwith of the wavelength table. The bandwith is
 * just computes as :
 * \li band(i) = (wlen(i+1) - wlen(i-1)) / 2.0
 *
 * @return
 * Always amdlibSUCCESS
 */
amdlibCOMPL_STAT amdlibComputeBandwidth(amdlibWAVEDATA *waveData)
{
    int l;
    for (l = 1; l < (amdlibNB_SPECTRAL_CHANNELS - 1); l++)
    {
        float *wlenTbl = waveData->wlen;
        float *widthTbl = waveData->bandwidth;
        if (wlenTbl[l] != 0)
        {
            if ((wlenTbl[l-1] != 0) && (wlenTbl[l+1] != 0))
            {
                widthTbl[l] = fabs((wlenTbl[l+1] - wlenTbl[l-1]) / 2.0);
            }
            else if ((wlenTbl[l-1] == 0) && (wlenTbl[l+1] != 0))
            {
                widthTbl[l] = fabs((wlenTbl[l+1] - wlenTbl[l]));
            }
            else if ((wlenTbl[l-1] != 0) && (wlenTbl[l+1] == 0))
            {
                widthTbl[l] = fabs((wlenTbl[l-1] - wlenTbl[l]));
            }
            else
            {
                widthTbl[l] = 0.0;
            }
        }
        else
        {
            widthTbl[l] = 0.0;
        }
    }  

    return amdlibSUCCESS;
}
/*___oOo__*/
